# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Stochastic gradient estimators.

These functions are meant to be used in conjunction with `StochasticTensor`
(`loss_fn` parameter) and `surrogate_loss`.

See Gradient Estimation Using Stochastic Computation Graphs
(http://arxiv.org/abs/1506.05254) by Schulman et al., eq. 1 and section 4, for
mathematical details.

## Score function estimator

The score function is an unbiased estimator of the gradient of `E_p(x)[f(x)]`,
where `f(x)` can be considered to be a "loss" term. It is computed as
`E_p(x)[f(x) grad(log p(x))]`. A constant `b`, referred to here as the
"baseline", can be subtracted from `f(x)` without affecting the expectation. The
term `(f(x) - b)` is referred to here as the "advantage".

Note that the methods defined in this module actually compute the integrand of
the score function, such that when taking the gradient, the true score function
is computed.

@@score_function
@@get_score_function_with_baseline
@@get_score_function_with_constant_baseline
@@get_score_function_with_advantage

## Baseline functions

Baselines reduce the variance of Monte Carlo estimate of an expectation. The
baseline for a stochastic node can be a function of all non-influenced nodes
(see section 4 of Schulman et al., linked above). Baselines are also known as
"control variates."

In the context of a MC estimate of `E_p(x)[f(x) - b]`, baseline functions have
the signature `(st, fx) => Tensor`, where `st` is a `StochasticTensor` backed by
the distribution `p(x)` and `fx` is the influenced loss.

@@get_mean_baseline

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import variable_scope as vs
from tensorflow.python.training import training
from tensorflow.python.util.all_util import make_all


def score_function(stochastic_tensor, value, loss, baseline=None,
                   name="ScoreFunction"):
  """Score function estimator.

  Computes the integrand of the score function with a baseline:
  `p.log_prob(value) * (loss - baseline)`.

  It will add a `stop_gradient` to the advantage `(loss - baseline)`.

  Args:
    stochastic_tensor: `StochasticTensor` p(x).
    value: `Tensor` x. Samples from p(x).
    loss: `Tensor`.
    baseline: `Tensor` broadcastable to `loss`.
    name: name to prepend ops with.

  Returns:
    `Tensor` `p.log_prob(x) * (loss - b)`. Taking the gradient yields the score
    function estimator.
  """
  with ops.name_scope(name, values=[value, loss, baseline]):
    value = ops.convert_to_tensor(value)
    loss = ops.convert_to_tensor(loss)
    if baseline is not None:
      baseline = ops.convert_to_tensor(baseline)
      advantage = loss - baseline
    else:
      advantage = loss

    advantage = array_ops.stop_gradient(advantage)
    return stochastic_tensor.distribution.log_prob(value) * advantage


def get_score_function_with_advantage(advantage_fn=None,
                                      name="ScoreFunctionWithAdvantage"):
  """Score function estimator with advantage function.

  Args:
    advantage_fn: callable that takes the `StochasticTensor` and the
      downstream `loss` and returns a `Tensor` advantage
      (e.g. `loss - baseline`).
    name: name to prepend ops with.

  Returns:
    Callable score function estimator that takes the `StochasticTensor`, the
    sampled `value`, and the downstream `loss`, and uses the provided advantage.
  """

  def score_function_with_advantage(stochastic_tensor, value, loss):
    with ops.name_scope(name, values=[value, loss]):
      advantage = advantage_fn(stochastic_tensor, loss)
      advantage = array_ops.stop_gradient(advantage)
      return stochastic_tensor.distribution.log_prob(value) * advantage

  return score_function_with_advantage


def get_score_function_with_constant_baseline(baseline, name="ScoreFunction"):
  """Score function estimator with constant baseline.

  Args:
    baseline: `Tensor` to be subtracted from loss.
    name: name to prepend ops with.

  Returns:
    Callable score function estimator that takes the `StochasticTensor`, the
    sampled `value`, and the downstream `loss`, and subtracts the provided
    `baseline` from the `loss`.
  """

  def score_function_with_constant_baseline(stochastic_tensor, value, loss):
    return score_function(stochastic_tensor, value, loss, baseline, name)

  return score_function_with_constant_baseline


def get_score_function_with_baseline(baseline_fn=None, name="ScoreFunction"):
  """Score function estimator with baseline function.

  Args:
    baseline_fn: callable that takes the `StochasticTensor` and the downstream
      `loss` and returns a `Tensor` baseline to be subtracted from the `loss`.
      If None, defaults to `get_mean_baseline`, which is an EMA of the loss.
    name: name to prepend ops with.

  Returns:
    Callable score function estimator that takes the `StochasticTensor`, the
    sampled `value`, and the downstream `loss`, and subtracts the provided
    `baseline` from the `loss`.
  """
  if baseline_fn is None:
    baseline_fn = get_mean_baseline()

  def score_function_with_baseline(stochastic_tensor, value, loss):
    with ops.name_scope(name):
      b = baseline_fn(stochastic_tensor, loss)
      return score_function(stochastic_tensor, value, loss, b)

  return score_function_with_baseline


def get_mean_baseline(ema_decay=0.99, name=None):
  """ExponentialMovingAverage baseline.

  Args:
    ema_decay: decay rate for the ExponentialMovingAverage.
    name: name for variable scope of the ExponentialMovingAverage.

  Returns:
    Callable baseline function that takes the `StochasticTensor` (unused) and
    the downstream `loss`, and returns an EMA of the loss.
  """

  def mean_baseline(_, loss):
    with vs.variable_scope(name, default_name="MeanBaseline"):
      reduced_loss = math_ops.reduce_mean(loss)

      ema = training.ExponentialMovingAverage(decay=ema_decay, zero_debias=True)
      update_op = ema.apply([reduced_loss])

      with ops.control_dependencies([update_op]):
        # Using `identity` causes an op to be added in this context, which
        # triggers the update. Removing the `identity` means nothing is updated.
        baseline = array_ops.identity(ema.average(reduced_loss))

      return baseline

  return mean_baseline


def get_vimco_advantage_fn(have_log_loss=False):
  """VIMCO (Variational Inference for Monte Carlo Objectives) baseline.

  Implements VIMCO baseline from the article of the same name:

  https://arxiv.org/pdf/1602.06725v2.pdf

  Given a `loss` tensor (containing non-negative probabilities or ratios),
  calculates the advantage VIMCO advantage via Eq. 9 of the above paper.

  The tensor `loss` should be shaped `[n, ...]`, with rank at least 1.  Here,
  the first axis is considered the single sampling dimension and `n` must
  be at least 2.  Specifically, the `StochasticTensor` is assumed to have
  used the `SampleValue(n)` value type with `n > 1`.

  Args:
    have_log_loss: Python `Boolean`.  If `True`, the loss is assumed to be the
      log loss.  If `False` (the default), it is assumed to be a nonnegative
      probability or probability ratio.

  Returns:
    Callable baseline function that takes the `StochasticTensor` (unused) and
    the downstream `loss`, and returns the VIMCO baseline for the loss.
  """
  def vimco_advantage_fn(_, loss, name=None):
    """Internal VIMCO function.

    Args:
      _: ignored `StochasticTensor`.
      loss: The loss `Tensor`.
      name: Python string, the name scope to use.

    Returns:
      The advantage `Tensor`.
    """
    with ops.name_scope(name, "VIMCOAdvantage", values=[loss]):
      loss = ops.convert_to_tensor(loss)
      loss_shape = loss.get_shape()
      loss_num_elements = loss_shape[0].value
      n = math_ops.cast(
          loss_num_elements or array_ops.shape(loss)[0], dtype=loss.dtype)

      if have_log_loss:
        log_loss = loss
      else:
        log_loss = math_ops.log(loss)

      # Calculate L_hat, Eq. (4) -- stably
      log_mean = math_ops.reduce_logsumexp(log_loss, [0]) - math_ops.log(n)

      # expand_dims: Expand shape [a, b, c] to [a, 1, b, c]
      log_loss_expanded = array_ops.expand_dims(log_loss, [1])

      # divide: log_loss_sub with shape [a, a, b, c], where
      #
      #  log_loss_sub[i] = log_loss - log_loss[i]
      #
      #       = [ log_loss[j] - log_loss[i] for rows j = 0 ... i - 1     ]
      #         [ zeros                                                  ]
      #         [ log_loss[j] - log_loss[i] for rows j = i + 1 ... a - 1 ]
      #
      log_loss_sub = log_loss - log_loss_expanded

      # reduce_sum: Sums each row across all the sub[i]'s; result is:
      #   reduce_sum[j] = (n - 1) * log_loss[j] - (sum_{i != j} loss[i])
      # divide by (n - 1) to get:
      #   geometric_reduction[j] =
      #     log_loss[j] - (sum_{i != j} log_loss[i]) / (n - 1)
      geometric_reduction = math_ops.reduce_sum(log_loss_sub, [0]) / (n - 1)

      # subtract this from the original log_loss to get the baseline:
      #   geometric_mean[j] = exp((sum_{i != j} log_loss[i]) / (n - 1))
      log_geometric_mean = log_loss - geometric_reduction

      ## Equation (9)

      # Calculate sum_{i != j} loss[i] -- via exp(reduce_logsumexp(.))
      # reduce_logsumexp: log-sum-exp each row across all the
      # -sub[i]'s, result is:
      #
      #  exp(reduce_logsumexp[j]) =
      #    1 + sum_{i != j} exp(log_loss[i] - log_loss[j])
      log_local_learning_reduction = math_ops.reduce_logsumexp(
          -log_loss_sub, [0])

      # convert local_learning_reduction to the sum-exp of the log-sum-exp
      #  (local_learning_reduction[j] - 1) * exp(log_loss[j])
      #    = sum_{i != j} exp(log_loss[i])
      local_learning_log_sum = (
          _logexpm1(log_local_learning_reduction) + log_loss)

      # Add (logaddexp) the local learning signals (Eq. 9)
      local_learning_signal = (
          math_ops.reduce_logsumexp(
              array_ops.stack((local_learning_log_sum, log_geometric_mean)),
              [0])
          - math_ops.log(n))

      advantage = log_mean - local_learning_signal

      return advantage

  return vimco_advantage_fn


def _logexpm1(x):
  """Stably calculate log(exp(x)-1)."""
  with ops.name_scope("logsumexp1"):
    eps = np.finfo(x.dtype.as_numpy_dtype).eps
    # Choose a small offset that makes gradient calculations stable for
    # float16, float32, and float64.
    safe_log = lambda y: math_ops.log(y + eps / 1e8)  # For gradient stability
    return array_ops.where(
        math_ops.abs(x) < eps,
        safe_log(x) + x/2 + x*x/24,  # small x approximation to log(expm1(x))
        safe_log(math_ops.exp(x) - 1))


__all__ = make_all(__name__)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""FisherFactor definitions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

import numpy as np
import six

from tensorflow.contrib.kfac.python.ops import utils
from tensorflow.python.framework import ops as tf_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import linalg_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import special_math_ops
from tensorflow.python.ops import variable_scope
from tensorflow.python.ops import variables
from tensorflow.python.training import moving_averages

# TODO(someone): come up with a better mechanism to set these constants
# externally. See b/67084987

# Whether to initialize covariance estimators at a zero matrix (or the identity
# matrix).
INIT_COVARIANCES_AT_ZERO = False

# Whether to zero-debias the moving averages.
ZERO_DEBIAS = False

# When the number of inverses requested from a FisherFactor exceeds this value,
# the inverses are computed using an eigenvalue decomposition.
EIGENVALUE_DECOMPOSITION_THRESHOLD = 2

# Numerical eigenvalues computed from covariance matrix estimates are clipped to
# be at least as large as this value before they are used to compute inverses or
# matrix powers. Must be nonnegative.
EIGENVALUE_CLIPPING_THRESHOLD = 0.0


def inverse_initializer(shape, dtype, partition_info=None):  # pylint: disable=unused-argument
  return array_ops.diag(array_ops.ones(shape[0], dtype))


def covariance_initializer(shape, dtype, partition_info=None):  # pylint: disable=unused-argument
  if INIT_COVARIANCES_AT_ZERO:
    return array_ops.diag(array_ops.zeros(shape[0], dtype))
  return array_ops.diag(array_ops.ones(shape[0], dtype))


def diagonal_covariance_initializer(shape, dtype, partition_info):  # pylint: disable=unused-argument
  if INIT_COVARIANCES_AT_ZERO:
    return array_ops.zeros(shape, dtype)
  return array_ops.ones(shape, dtype)


def _compute_cov(tensor, normalizer=None):
  """Compute the empirical second moment of the rows of a 2D Tensor.

  This function is meant to be applied to random matrices for which the true row
  mean is zero, so that the true second moment equals the true covariance.

  Args:
    tensor: A 2D Tensor.
    normalizer: optional scalar for the estimator (by default, the normalizer is
        the number of rows of tensor).

  Returns:
    A square 2D Tensor with as many rows/cols as the number of input columns.
  """
  if normalizer is None:
    normalizer = array_ops.shape(tensor)[0]
  cov = (math_ops.matmul(tensor, tensor, transpose_a=True) / math_ops.cast(
      normalizer, tensor.dtype))
  return (cov + array_ops.transpose(cov)) / math_ops.cast(2, cov.dtype)


def _append_homog(tensor):
  """Appends a homogeneous coordinate to the last dimension of a Tensor.

  Args:
    tensor: A Tensor.

  Returns:
    A Tensor identical to the input but one larger in the last dimension.  The
    new entries are filled with ones.
  """
  rank = len(tensor.shape.as_list())
  shape = array_ops.concat([array_ops.shape(tensor)[:-1], [1]], axis=0)
  ones = array_ops.ones(shape, dtype=tensor.dtype)
  return array_ops.concat([tensor, ones], axis=rank-1)


def scope_string_from_params(params):
  """Builds a variable scope string name from the given parameters.

  Supported parameters are:
    * tensors
    * booleans
    * ints
    * strings
    * depth-1 tuples/lists of ints
    * any depth tuples/lists of tensors
  Other parameter types will throw an error.

  Args:
    params: A parameter or list of parameters.

  Returns:
    A string to use for the variable scope.

  Raises:
    ValueError: if params includes an unsupported type.
  """
  params = params if isinstance(params, (tuple, list)) else (params,)

  name_parts = []
  for param in params:
    if isinstance(param, (tuple, list)):
      if all([isinstance(p, int) for p in param]):
        name_parts.append("-".join([str(p) for p in param]))
      else:
        name_parts.append(scope_string_from_name(param))
    elif isinstance(param, (str, int, bool)):
      name_parts.append(str(param))
    elif isinstance(param, (tf_ops.Tensor, variables.Variable)):
      name_parts.append(scope_string_from_name(param))
    else:
      raise ValueError(
          "Encountered an unsupported param type {}".format(type(param)))
  return "_".join(name_parts)


def scope_string_from_name(tensor):
  if isinstance(tensor, (tuple, list)):
    return "__".join([scope_string_from_name(t) for t in tensor])
  # "gradients/add_4_grad/Reshape:0" -> "gradients_add_4_grad_Reshape"
  return tensor.name.split(":")[0].replace("/", "_")


def scalar_or_tensor_to_string(val):
  return repr(val) if np.isscalar(val) else scope_string_from_name(val)


@six.add_metaclass(abc.ABCMeta)
class FisherFactor(object):
  """Base class for objects modeling factors of approximate Fisher blocks.

     Note that for blocks that aren't based on approximations, a 'factor' can
     be the entire block itself, as is the case for the diagonal and full
     representations.

     Subclasses must implement the _compute_new_cov method, and the _var_scope
     and _cov_shape properties.
  """

  def __init__(self):
    self.instantiate_covariance()

  @abc.abstractproperty
  def _var_scope(self):
    pass

  @abc.abstractproperty
  def _cov_shape(self):
    """The shape of the cov matrix."""
    pass

  @abc.abstractproperty
  def _num_sources(self):
    """The number of things to sum over when computing cov.

    The default make_covariance_update_op function will call _compute_new_cov
    with indices ranging from 0 to _num_sources-1. The typical situation is
    where the factor wants to sum the statistics it computes over multiple
    backpropped "gradients" (typically passed in via "tensors" or
    "outputs_grads" arguments).
    """
    pass

  @property
  def _cov_initializer(self):
    return covariance_initializer

  def instantiate_covariance(self):
    """Instantiates the covariance Variable as the instance member _cov."""
    with variable_scope.variable_scope(self._var_scope):
      self._cov = variable_scope.get_variable(
          "cov",
          initializer=self._cov_initializer,
          shape=self._cov_shape,
          trainable=False)

  @abc.abstractmethod
  def _compute_new_cov(self, idx=0):
    pass

  def make_covariance_update_op(self, ema_decay):
    """Constructs and returns the covariance update Op.

    Args:
      ema_decay: The exponential moving average decay (float or Tensor).
    Returns:
      An Op for updating the covariance Variable referenced by _cov.
    """
    new_cov = math_ops.add_n(
        tuple(self._compute_new_cov(idx) for idx in range(self._num_sources)))

    return moving_averages.assign_moving_average(
        self._cov, new_cov, ema_decay, zero_debias=ZERO_DEBIAS)

  def make_inverse_update_ops(self):
    """Create and return update ops corresponding to registered computations."""
    return []

  def get_cov(self):
    return self._cov


class InverseProvidingFactor(FisherFactor):
  """Base class for FisherFactors that maintain inverses, powers, etc of _cov.

  Assumes that the _cov property is a square PSD matrix.

  Subclasses must implement the _compute_new_cov method, and the _var_scope and
  _cov_shape properties.
  """

  def __init__(self):
    self._inverses_by_damping = {}
    self._matpower_by_exp_and_damping = {}
    self._eigendecomp = None

    super(InverseProvidingFactor, self).__init__()

  def register_damped_inverse(self, damping):
    """Registers a damped inverse needed by a FisherBlock.

    Args:
      damping: The damping value (float or Tensor) for this factor.
    """
    if damping not in self._inverses_by_damping:
      damping_string = scalar_or_tensor_to_string(damping)
      with variable_scope.variable_scope(self._var_scope):
        inv = variable_scope.get_variable(
            "inv_damp{}".format(damping_string),
            initializer=inverse_initializer,
            shape=self._cov_shape,
            trainable=False)
      self._inverses_by_damping[damping] = inv

  def register_matpower(self, exp, damping):
    """Registers a matrix power needed by a FisherBlock.

    Args:
      exp: The exponent (float or Tensor) to raise the matrix to.
      damping: The damping value (float or Tensor).
    """
    if (exp, damping) not in self._matpower_by_exp_and_damping:
      exp_string = scalar_or_tensor_to_string(exp)
      damping_string = scalar_or_tensor_to_string(damping)
      with variable_scope.variable_scope(self._var_scope):
        matpower = variable_scope.get_variable(
            "matpower_exp{}_damp{}".format(exp_string, damping_string),
            initializer=inverse_initializer,
            shape=self._cov_shape,
            trainable=False)
      self._matpower_by_exp_and_damping[(exp, damping)] = matpower

  def register_eigendecomp(self):
    """Registers that an eigendecomposition is needed by a FisherBlock."""
    if not self._eigendecomp:
      self._eigendecomp = linalg_ops.self_adjoint_eig(self._cov)

  def make_inverse_update_ops(self):
    """Create and return update ops corresponding to registered computations."""
    ops = super(InverseProvidingFactor, self).make_inverse_update_ops()

    num_inverses = len(self._inverses_by_damping)
    matrix_power_registered = bool(self._matpower_by_exp_and_damping)
    use_eig = (self._eigendecomp or matrix_power_registered or
               num_inverses >= EIGENVALUE_DECOMPOSITION_THRESHOLD)

    if use_eig:
      self.register_eigendecomp()  # ensures self._eigendecomp is set
      eigenvalues, eigenvectors = self._eigendecomp  # pylint: disable=unpacking-non-sequence

      # The matrix self._cov is positive semidefinite by construction, but the
      # numerical eigenvalues could be negative due to numerical errors, so here
      # we clip them to be at least EIGENVALUE_CLIPPING_THRESHOLD.
      clipped_eigenvalues = math_ops.maximum(eigenvalues,
                                             EIGENVALUE_CLIPPING_THRESHOLD)

      for damping, inv in self._inverses_by_damping.items():
        ops.append(
            inv.assign(
                math_ops.matmul(eigenvectors / (clipped_eigenvalues + damping),
                                array_ops.transpose(eigenvectors))))

      for (exp, damping), matpower in self._matpower_by_exp_and_damping.items():
        ops.append(
            matpower.assign(
                math_ops.matmul(eigenvectors * (clipped_eigenvalues + damping)**
                                exp, array_ops.transpose(eigenvectors))))
    else:
      for damping, inv in self._inverses_by_damping.items():
        ops.append(inv.assign(utils.posdef_inv(self._cov, damping)))

    return ops

  def get_inverse(self, damping):
    return self._inverses_by_damping[damping]

  def get_matpower(self, exp, damping):
    return self._matpower_by_exp_and_damping[(exp, damping)]

  def get_eigendecomp(self):
    return self._eigendecomp


class FullFactor(InverseProvidingFactor):
  """FisherFactor for a full matrix representation of the Fisher of a parameter.

  Note that this uses the naive "square the sum estimator", and so is applicable
  to any type of parameter in principle, but has very high variance.
  """

  def __init__(self, params_grads, batch_size):
    self._batch_size = batch_size
    self._orig_params_grads_name = scope_string_from_params(
        [params_grads, self._batch_size])
    self._params_grads_flat = tuple(
        utils.tensors_to_column(params_grad) for params_grad in params_grads)
    super(FullFactor, self).__init__()

  @property
  def _var_scope(self):
    return "ff_full/" + self._orig_params_grads_name

  @property
  def _cov_shape(self):
    size = self._params_grads_flat[0].shape[0]
    return [size, size]

  @property
  def _num_sources(self):
    return len(self._params_grads_flat)

  def _compute_new_cov(self, idx=0):
    # This will be a very basic rank 1 estimate
    return ((self._params_grads_flat[idx] * array_ops.transpose(
        self._params_grads_flat[idx])) / math_ops.cast(
            self._batch_size, self._params_grads_flat[idx].dtype))


class DiagonalFactor(FisherFactor):
  """A base class for FisherFactors that use diagonal approximations."""

  def __init__(self):
    super(DiagonalFactor, self).__init__()

  @property
  def _cov_initializer(self):
    return diagonal_covariance_initializer


class NaiveDiagonalFactor(DiagonalFactor):
  """FisherFactor for a diagonal approximation of any type of param's Fisher.

  Note that this uses the naive "square the sum estimator", and so is applicable
  to any type of parameter in principle, but has very high variance.
  """

  def __init__(self, params_grads, batch_size):
    self._batch_size = batch_size
    self._params_grads = tuple(
        utils.tensors_to_column(params_grad) for params_grad in params_grads)
    self._orig_params_grads_name = scope_string_from_params(
        [self._params_grads, self._batch_size])
    super(NaiveDiagonalFactor, self).__init__()

  @property
  def _var_scope(self):
    return "ff_naivediag/" + self._orig_params_grads_name

  @property
  def _cov_shape(self):
    return self._params_grads[0].shape

  @property
  def _num_sources(self):
    return len(self._params_grads)

  def _compute_new_cov(self, idx=0):
    return (math_ops.square(self._params_grads[idx]) / math_ops.cast(
        self._batch_size, self._params_grads[idx].dtype))


class FullyConnectedDiagonalFactor(DiagonalFactor):
  """FisherFactor for a diagonal approx of a fully-connected layer's Fisher."""

  # TODO(jamesmartens): add units tests for this class

  def __init__(self, inputs, outputs_grads, has_bias=False):
    self._outputs_grads = outputs_grads
    self._batch_size = array_ops.shape(inputs)[0]
    self._orig_tensors_name = scope_string_from_params((inputs,) +
                                                       tuple(outputs_grads))

    # Note that we precompute the required operations on the inputs since the
    # inputs don't change with the 'idx' argument to _compute_new_cov.  (Only
    # the target entry of _outputs_grads changes with idx.)
    if has_bias:
      inputs = _append_homog(inputs)
    self._squared_inputs = math_ops.square(inputs)

    super(FullyConnectedDiagonalFactor, self).__init__()

  @property
  def _var_scope(self):
    return "ff_diagfc/" + self._orig_tensors_name

  @property
  def _cov_shape(self):
    return [self._squared_inputs.shape[1], self._outputs_grads[0].shape[1]]

  @property
  def _num_sources(self):
    return len(self._outputs_grads)

  def _compute_new_cov(self, idx=0):
    # The well-known special formula that uses the fact that the entry-wise
    # square of an outer product is the outer-product of the entry-wise squares.
    # The gradient is the outer product of the input and the output gradients,
    # so we just square both and then take their outer-product.
    new_cov = math_ops.matmul(
        self._squared_inputs,
        math_ops.square(self._outputs_grads[idx]),
        transpose_a=True)
    new_cov /= math_ops.cast(self._batch_size, new_cov.dtype)
    return new_cov


class ConvDiagonalFactor(DiagonalFactor):
  """FisherFactor for a diagonal approx of a convolutional layer's Fisher."""

  # TODO(jamesmartens): add units tests for this class

  def __init__(self, inputs, outputs_grads, filter_shape, strides, padding,
               has_bias=False):
    """Creates a ConvDiagonalFactor object.

    Args:
      inputs: Tensor of shape [batch_size, height, width, in_channels].
        Input activations to this layer.
      outputs_grads: Tensor of shape [batch_size, height, width, out_channels].
        Per-example gradients to the loss with respect to the layer's output
        preactivations.
      filter_shape: Tuple of 4 ints: (kernel_height, kernel_width, in_channels,
        out_channels). Represents shape of kernel used in this layer.
      strides: The stride size in this layer (1-D Tensor of length 4).
      padding: The padding in this layer (1-D of Tensor length 4).
      has_bias: Python bool. If True, the layer is assumed to have a bias
        parameter in addition to its filter parameter.
    """
    self._filter_shape = filter_shape
    self._has_bias = has_bias
    self._outputs_grads = outputs_grads

    self._orig_tensors_name = scope_string_from_name((inputs,)
                                                     + tuple(outputs_grads))

    # Note that we precompute the required operations on the inputs since the
    # inputs don't change with the 'idx' argument to _compute_new_cov.  (Only
    # the target entry of _outputs_grads changes with idx.)
    filter_height, filter_width, _, _ = self._filter_shape
    patches = array_ops.extract_image_patches(
        inputs,
        ksizes=[1, filter_height, filter_width, 1],
        strides=strides,
        rates=[1, 1, 1, 1],
        padding=padding)

    if has_bias:
      patches = _append_homog(patches)

    self._patches = patches

    super(ConvDiagonalFactor, self).__init__()

  @property
  def _var_scope(self):
    return "ff_convdiag/" + self._orig_tensors_name

  @property
  def _cov_shape(self):
    filter_height, filter_width, in_channels, out_channels = self._filter_shape
    return [filter_height * filter_width * in_channels + self._has_bias,
            out_channels]

  @property
  def _num_sources(self):
    return len(self._outputs_grads)

  def _compute_new_cov(self, idx=0):
    outputs_grad = self._outputs_grads[idx]
    batch_size = array_ops.shape(self._patches)[0]

    new_cov = self._convdiag_sum_of_squares(self._patches, outputs_grad)
    new_cov /= math_ops.cast(batch_size, new_cov.dtype)

    return new_cov

  def _convdiag_sum_of_squares(self, patches, outputs_grad):
    # This computes the sum of the squares of the per-training-case "gradients".
    # It does this simply by computing a giant tensor containing all of these,
    # doing an entry-wise square, and them summing along the batch dimension.
    case_wise_gradients = special_math_ops.einsum("bijk,bijl->bkl", patches,
                                                  outputs_grad)
    return math_ops.reduce_sum(math_ops.square(case_wise_gradients), axis=0)


class FullyConnectedKroneckerFactor(InverseProvidingFactor):
  """Kronecker factor for the input or output side of a fully-connected layer.
  """

  def __init__(self, tensors, has_bias=False):
    # The tensor argument is either a tensor of input activations or a tensor of
    # output pre-activation gradients.
    self._has_bias = has_bias
    self._tensors = tensors
    super(FullyConnectedKroneckerFactor, self).__init__()

  @property
  def _var_scope(self):
    return "ff_fckron/" + scope_string_from_params(
        [self._tensors, self._has_bias])

  @property
  def _cov_shape(self):
    size = self._tensors[0].shape[1] + self._has_bias
    return [size, size]

  @property
  def _num_sources(self):
    return len(self._tensors)

  def _compute_new_cov(self, idx=0):
    tensor = self._tensors[idx]
    if self._has_bias:
      tensor = _append_homog(tensor)
    return _compute_cov(tensor)


class ConvInputKroneckerFactor(InverseProvidingFactor):
  """Kronecker factor for the input side of a convolutional layer."""

  def __init__(self, inputs, filter_shape, strides, padding, has_bias=False):
    self._filter_shape = filter_shape
    self._strides = strides
    self._padding = padding
    self._has_bias = has_bias
    self._inputs = inputs
    super(ConvInputKroneckerFactor, self).__init__()

  @property
  def _var_scope(self):
    return "ff_convinkron/" + scope_string_from_params([
        self._inputs, self._filter_shape, self._strides, self._padding,
        self._has_bias
    ])

  @property
  def _cov_shape(self):
    filter_height, filter_width, in_channels, _ = self._filter_shape
    size = filter_height * filter_width * in_channels + self._has_bias
    return [size, size]

  @property
  def _num_sources(self):
    return 1

  def _compute_new_cov(self, idx=0):
    if idx != 0:
      raise ValueError("ConvInputKroneckerFactor only supports idx = 0")

    # TODO(jamesmartens): factor this patches stuff out into a utility function
    filter_height, filter_width, in_channels, _ = self._filter_shape
    patches = array_ops.extract_image_patches(
        self._inputs,
        ksizes=[1, filter_height, filter_width, 1],
        strides=self._strides,
        rates=[1, 1, 1, 1],
        padding=self._padding)

    flatten_size = (filter_height * filter_width * in_channels)
    patches_flat = array_ops.reshape(patches, [-1, flatten_size])

    if self._has_bias:
      patches_flat = _append_homog(patches_flat)

    return _compute_cov(patches_flat)


class ConvOutputKroneckerFactor(InverseProvidingFactor):
  """Kronecker factor for the output side of a convolutional layer."""

  def __init__(self, outputs_grads):
    self._out_channels = outputs_grads[0].shape.as_list()[3]
    self._outputs_grads = outputs_grads
    super(ConvOutputKroneckerFactor, self).__init__()

  @property
  def _var_scope(self):
    return "ff_convoutkron/" + scope_string_from_params(self._outputs_grads)

  @property
  def _cov_shape(self):
    size = self._out_channels
    return [size, size]

  @property
  def _num_sources(self):
    return len(self._outputs_grads)

  def _compute_new_cov(self, idx=0):
    reshaped_tensor = array_ops.reshape(self._outputs_grads[idx],
                                        [-1, self._out_channels])
    return _compute_cov(reshaped_tensor)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Library to compute order of computations in a graph.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections


class GraphDefHelper(object):
  """Helper class to collect node names and definitions.

  Example:
    b = GraphDefHelper(graph_def)
    # Prints node that produces given output.
    print b.output_of['conv/foo/bar']
  """

  def __init__(self, gd):
    self.output_of = {}
    for each in gd.node:
      self.output_of[each.name] = each


# pylint: disable=invalid-name
_NodeEntry = collections.namedtuple('NodeEntry', field_names=['order', 'node'])


def _get_computed_nodes(g, output, seen):
  """Traverses the graph in topological order.

  Args:
    g: GraphDefHelper object.
    output: current node.
    seen: map of nodes we've already traversed.
  Returns:
    order in topological sort for 'output'.
  """
  if output in seen:
    return seen[output].order
  node_def = g.output_of.get(output, None)
  if node_def is None:
    seen[output] = _NodeEntry(0, None)
    return 0

  r = 0
  for each in node_def.input:
    # Parses name of input node.
    if each.startswith('^'):
      each = each[1:]
    each = each.split(':')[0]
    # Recursively computes ordering.
    new_v = _get_computed_nodes(g, each, seen)
    r = max(r, new_v + 1)

  seen[output] = _NodeEntry(r, node_def)

  return seen[output].order


def get_compute_order(graph_def):
  """Computes order of computation for a given graph.

  Args:
    graph_def: GraphDef object.
  Returns:
    map: name -> {order, node}
  """
  helper = GraphDefHelper(graph_def)
  seen = collections.defaultdict(_NodeEntry)
  for each in graph_def.node:
    _get_computed_nodes(helper, each.name, seen)
  return seen
